.error-page .error-message-container {
    margin-left: auto;
    margin-right: auto;
    padding-top: 40px;
    max-width: 1400px;
    width: 87%;
}
 .error-page .error-message {
    min-height: 100vh;
    background: var(--ifm-blockquote-color);
}
 .error-page .error-message span {
    color: var(--ifm-color-primary);
    font-size: 8.8em;
    font-weight: 700;
    display: inline-block;
    margin-top: 10vh;
    text-align: center;
    display: block;
}
 .error-page .error-message p {
    margin-top: 50px;
    font-size: 1.6em;
    text-align: center;
}

.error-page .error-message a {
   margin-bottom: 50px;
   font-size: 1.6em;
   text-align: center;
   display: block;
   text-decoration: underline;
}
